import logging
from binascii import hexlify
import mems.protocol.Rosco

class MemsClient(object):
    def __init__(self, interface):
        self._logger = logging.getLogger('mems.client')

        self._rosco = mems.protocol.Rosco.Rosco()
        self._interface = interface
        self._initialized = False

    @property
    def responses(self):
        return self._interface.recieve_queue

    @property
    def connected(self):
        return self._interface.connected

    @property
    def has_data_to_send(self):
        return self._interface.has_data_to_send()

    def connect(self):
        if self._interface.connected == False:
            self._interface.connect()
            self._logger.info('connected')

    def disconnect(self):
        if self._interface.connected == True:
            self._interface.disconnect()
            self._logger.info('disconnected')

    def hex_to_string(self, code):
        request_code = hexlify(code)
        return str(request_code.decode("utf-8")).upper()

    def send_command(self, command, data=bytearray()):
        # get the command code for the command
        command_code = self._rosco.get_command_code(command)

        mems_command = bytearray()
        mems_command.extend(command_code)

        # add the data payload if required
        if data is not None:
            if len(data) > 0:
                mems_command.extend(data)

        return self._send_data(mems_command)

    def _send_data(self, mems_command):
        self._logger.debug('sending command %s to MEMS',  self.hex_to_string(mems_command))

        expected_result = self._rosco.get_expected_result(mems_command)
        expected_bytes = len(expected_result)

        # send the command
        self._interface.send(mems_command, expected_bytes)

        self._logger.debug('waiting for response from MEMS..')
        # read the response
        # or wait until the read timeout has expired
        while self._interface.waiting_for_mems_response:
            pass

        response = self._interface.read()
        self._logger.info('%s response from MEMS', self.hex_to_string(response))

        return response



import logging
import time
import subprocess
import threading
import mems.protocol.Rosco
from binascii import hexlify

from mems.interfaces.SerialInterface import SerialInterface

class MemsEmulator(SerialInterface):
    def __init__(self, connection):
        self._rosco = mems.protocol.Rosco.Rosco()

        self._serial_port_loopback = threading.Thread(target=self._create_virtual_ports)
        self._serial_port_loopback.start()
        time.sleep(1)

        super().__init__(connection)
        self._logger = logging.getLogger('mems.emulator.interface')

    def disconnect(self):
        yield super().disconnect()
        self._remove_virtal_ports()
        time.sleep(1)

    def receive_loop_hook(self, data):
        super().receive_loop_hook(data)
        response = self._get_response(data)
        self._logger.debug('mems response %s', self.hex_to_string(data))
        self.send(response)

    def hex_to_string(self, code):
        request_code = hexlify(code)
        return str(request_code.decode("utf-8")).upper()

    def _get_response(self, code):
        return self._rosco.get_response(code)

    def _create_virtual_ports(self):
        cmd = ['socat', '-d', '-d', 'pty,link=ttycodereader,raw,echo=0', 'pty,link=ttyecu,raw,echo=0']
        try:
            self.proc = subprocess.Popen(cmd)
        except Exception as ex:
            print (ex)

    def _remove_virtal_ports(self):
        # kill the virtual ports
        self.proc.kill()


# ROSCO - Rover Communication Protocol
import os
from binascii import hexlify

class Rosco(object):
    def __init__(self):
        self._version   = 'MNE101170'
        self._connected = False

        self._versions = {
                'MNE101070' : b'\xd0\x99\x00\x02\x03',
                'MNE101170' : b'\xd0\x99\x00\x03\x03'
            }

        #
        # The tables below describe the known fields in the data frames
        # that are sent by the ECU in responses to commands 0x7D and 0x80, respectively.
        # Multi-byte fields are sent in big-endian format.
        #

        self._dataframes = [
                {'command' : b'\x7d', 'fields' : ['command','dataframe_size','','throttle_angle','', '','lambda_voltage','lambda_frequency','lambda_duty_cycle','lambda_status','loop_indicator','long_term_trim','short_term_trim','carbon_canister_purge_valve_duty_cycle','','idle_base_position','','','','','idle_error','','','','','','','','','','','','']},
                {'command' : b'\x80', 'fields' : ['command','dataframe_size','engine_speed_low_byte','engine_speed_high_byte','coolant_temperature','ambient_temperature','intake_air_temperature','fuel_temperature','map_sensor','battery_voltage','throttle_pot_voltage','idle_switch','','park_neutral_switch','coolant_temp_inlet_air_temp_sensor_fault','fuel_pump_throttle_pot_circuit_fault','','','','idle_air_contol_position','idle_speed_deviation_low_byte','idle_speed_deviation_high_byte','','ignition_advance','coil_time_low_byte','coil_time_high_byte','','','']}
            ]

        self._commands = [
                {'open_fuel_pump_relay' : b'\x01'},
                {'open_ptc_relay'       : b'\x02'},
                {'open_aircond_relay'   : b'\x03'},
                {'close_purge_valve'    : b'\x08'},
                {'open_heater_relay'    : b'\x09'},
                {'close_fuel_pump_relay': b'\x11'},
                {'close_ptc_relay'      : b'\x12'},
                {'close_aircond_relay'  : b'\x13'},
                {'open_purge_valve'     : b'\x18'},
                {'close_heater_relay'   : b'\x19'},
                {'close_fan_relay'      : b'\x1e'},
                {'request_data_frame_a' : b'\x7d'},
                {'inc_fuel_trim'        : b'\x79'},
                {'dec_fuel_trim'        : b'\x7a'},
                {'inc_fuel_trim_alt'    : b'\x7b'},
                {'dec_fuel_trim_alt'    : b'\x7c'},
                {'request_data_frame_b' : b'\x80'},
                {'inc_idle_decay'       : b'\x89'},
                {'dec_idle_decay'       : b'\x8a'},
                {'inc_idle_speed'       : b'\x91'},
                {'dec_idle_speed'       : b'\x92'},
                {'inc_ignition_advance' : b'\x93'},
                {'dec_ignition_advance' : b'\x94'},
                {'clear_fault_codes'    : b'\xcc'},
                {'heartbeat'            : b'\xf4'},
                {'actuate_fuel_injector': b'\xf7'},
                {'fire_ignition_coil'   : b'\xf8'},
                {'current_iac_position' : b'\xfb'},
                {'open_iac_by_one_step' : b'\xfd'},
                {'close_iac_by_one_step': b'\xfe'},
            ]

        self._response = [
            {'command': b'\x00', 'response': b'\x00\x00'},
            {'command': b'\x1d', 'response': b'\x1d'},
            {'command': b'\x1e', 'response': b'\x1e'},
            {'command': b'\x7d', 'response': b'\x7d\x20\x10\x13\xFF\x92\x00\x57\xFF\xFF\x01\x00\x95\x64\x00\xFF\x5D\xFF\xFF\x30\x80\x80\x03\xFF\x1B\xC0\x21\x40\x29\x40\x37\x80\x17'},
            {'command': b'\x7e', 'response': b'\x7e\x08'},
            {'command': b'\x7f', 'response': b'\x7f\x05'},
            {'command': b'\x75', 'response': b'\x75'},
            {'command': b'\x80', 'response': b'\x80\x1C\x04\xAE\x4D\xFF\x4B\xFF\x25\x73\x26\x00\x10\x01\x00\x00\x00\x19\x90\x6C\x00\x14\x00\x47\x08\xB2\x10\x00\x00'},
            {'command': b'\x82', 'response': b'\x82\x09\x9E\x1D\x00\x00\x60\x05\xFF\xFF'},
            {'command': b'\xca', 'response': b'\xca'},
            {'command': b'\xcd', 'response': b'\xcc\x00'},
            {'command': b'\xcd', 'response': b'\xcd\x01'},
            {'command': b'\xd0', 'response': b'\xd0\x99\x00\x03\x03'},
            {'command': b'\xd1', 'response': b'\xd1\x41\x42\x4E\x4D\x50\x30\x30\x33\x99\x00\x03\x03\x41\x42\x4E\x4D\x50\x30\x30\x33\x99\x00\x03\x03\x41\x42\x4E\x4D\x50\x30\x30\x33\x99\x00\x03\x03'},
            {'command': b'\xe7', 'response': b'\xe7\x02'},
            {'command': b'\xe8', 'response': b'\xe8\x05\x26\x01\x00\x01'},
            {'command': b'\xf4', 'response': b'\xf4\x00'},
        ]

    def hex_to_string(self, code):
        request_code = hexlify(code)
        return str(request_code.decode("utf-8")).upper()

    def get_expected_result(self, code):
        response = self.get_response(code)
        return response

    def get_response(self, code):
        for response in self._response:
            if code == response['command']:
                return response['response']

    def get_command_code(self, command_name):
         for c in self._commands:
            if command_name in c:
                return c[command_name]

         return command_name

    def get_version(self, response_code):
        for key, value in self._versions.items():
            if response_code == value:
                return key

    def get_dataframe(self, command_name):
        code = self.get_command_code(command_name)

        for dataframe in self._dataframes:
            if code == dataframe['command']:
                return dataframe['fields']

    @property
    def initialization_sequence(self):
        return [
                   {'tx': b'\xca', 'response': b'\xca'},
                   {'tx': b'\x75', 'response': b'\x75'},
                   {'tx': b'\xd0', 'response': self._version}
               ]




from mems.MemsClient import MemsClient
from mems.MemsEmulator import MemsEmulator
from mems.MemsReader import MemsReader
from mems.interfaces.MemsSerialInterface import MemsSerialInterface
from mems.interfaces.SerialInterface import SerialInterface
from mems.protocol.Rosco import Rosco


import logging
from mems.interfaces.SerialInterface import SerialInterface

class MemsSerialInterface(SerialInterface):
    def __init__(self, connection):
        super().__init__(connection)
        self._logger = logging.getLogger('mems.serial.interface')
        self._waiting_for_mems_response = False

    @property
    def waiting_for_mems_response(self):
        return self._waiting_for_mems_response

    def connect(self):
        super().connect()
        self._logger.debug('connected')

    def disconnect(self):
        super().disconnect()
        self._logger.debug('disconnected')

    def send(self, data, expected_bytes):
        self._waiting_for_mems_response = True
        super().send(data, expected_bytes)
        self._logger.debug('sent %s to MEMS', str(data))

    def receive_loop_hook(self, data):
        self._waiting_for_mems_response = False
        super().receive_loop_hook(data)

    def read(self):
        data = bytearray()

        while self._receive_queue.qsize() > 0:
            data.extend( super().read() )

        self._logger.debug('read %s from MEMS', str(data))
        return data



import logging
import serial
from binascii import hexlify
import mems.protocol.Rosco
from mems import MemsEmulator
from mems.interfaces.MemsSerialInterface import MemsSerialInterface

class MemsReader(object):
    def __init__(self, serial_port, is_emulator=False):
        self._logger = logging.getLogger('mems.reader')
        self._client = None

        mems_connection = {
            'port' : '/dev/null',
            'baudrate': 9600,
            'bytesize': serial.EIGHTBITS,
            'stopbits': serial.STOPBITS_ONE,
            'parity': serial.PARITY_NONE,
            'xonxoff': False,
            'rtscts': False,
            'dsrdtr': False,
            'write_timeout': 1,
            'timeout': 0
        }

        # create the serial interface to connect to MEMS ECU
        mems_connection['port'] = serial_port

        if is_emulator:
            self._client = mems.MemsEmulator(mems_connection)
        else:
            interface = MemsSerialInterface(mems_connection)
            self._client = mems.MemsClient(interface)

    @property
    def mems_interface(self):
        return self._client

    def connect(self):
        self._client.connect()

    def disconnect(self):
        self._client.disconnect()


from setuptools import setup

setup(name='memsulator',
      version='1.0.2',
      description='Rover MEMS 1.6 Serial Interface and Emulator',
      url='http://github.com/andrewdjackson/memsulator',
      author='Andrew Jackson',
      author_email='andrew.d.jackson@gmail.com',
      license='MIT',
      packages=['mems','mems.interfaces','mems.protocol'],
      zip_safe=False)


import queue
from binascii import hexlify

import serial
import threading
import logging
import time

class SerialInterface():
    def __init__(self, connection):
        self._logger = logging.getLogger('serial.interface')

        self._send_queue = queue.Queue()
        self._receive_queue = queue.Queue()
        self._connected = False

        # to use serial_for_url the port and the settings need to be separated out
        # from the connection dictionary and then removed to prevent a duplicate
        # 'port' error
        self._port = connection['port']
        self._connection = connection
        del self._connection['port']

    @property
    def receive_queue(self):
        return self._receive_queue

    @property
    def connected(self):
        return self._connected

    def connect(self):
        # create a thread to listen for commands
        self._connected = True
        self._listener = threading.Thread(target=self._send_receive_loop)

        # create the virtual ports for emulate MEMS
        # self._create_virtual_ports()

        # open port for listening
        self.serial = serial.serial_for_url(self._port, **self._connection)
        self.serial.flush()
        # start listening for commands and responses
        self._listener.start()

        self._logger.debug('connected')

    def disconnect(self):
        while self.has_data_to_send():
            pass

        # stop listening
        self._connected = False

        return True

    def send(self, data, expected_result_length = 0):
        # add data to the send buffer
        self._send_queue.put({'data':data, 'expected_result_length':expected_result_length})

    def read(self):
        # read data from the read buffer
        if self.has_received_data():
            data = self._receive_queue.get()

            self._logger.debug('read %s from receive queue', self.hex_to_string(data))
            return data
        else:
            return None

    def _send_receive_loop(self):
        while self._connected:
            expected_result_length = 0

            if self.has_data_to_send():
                send_data = self._send_queue.get()
                data = send_data['data']
                expected_result_length = send_data['expected_result_length']

                self.serial.write(data)
                self.serial.flush()
                self._logger.debug('sent %s', self.hex_to_string(data))

            while self.serial.in_waiting:
                # determine the expected length to be received
                if (expected_result_length > 0):
                    data = self.serial.read(expected_result_length)
                else:
                    data = self.serial.read()

                self.serial.flush()
                self._logger.debug('read %s', self.hex_to_string(data))
                self._receive_queue.put(data)

                # this allows a response to be generated based on the received
                # data. This function can be overridden for custom processing.
                self.receive_loop_hook(data)

        # sleep for 50ms to yield time to the cpu without interfering with
        # the serial communications
        time.sleep(0.5)

    def hex_to_string(self, code):
        request_code = hexlify(code)
        return str(request_code.decode("utf-8")).upper()

    def receive_loop_hook(self, data):
        pass

    def has_data_to_send(self):
        return not self._send_queue.empty()

    def has_received_data(self):
        return not self._receive_queue.empty()
